from functools import reduce
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from .types import ColumnSpec, IndexFormatter
import inspect

__all__ = ["Tree"]


class Tree:
    def __init__(self, parent: Optional["Tree"], path: str) -> None:
        self.parent: Union[Tree, None] = parent
        self.depth: int = 0
        if parent and parent.depth is not None:
            self.depth = parent.depth + 1

        self.path: str = path
        self.children: List[Tree] = []
        self.tree_spec: ColumnSpec = {}
        self.accepts_index_arg = False

    @property
    def name(self) -> Any:
        if self.tree_spec.get("name"):
            return self.tree_spec.get("name")
        return self.path

    def get_value(self, value: Any, idx: Union[int, None] = None) -> Any:
        formatter = self.tree_spec.get("formatter")
        if formatter:
            if isinstance(formatter, IndexFormatter) and idx is not None and self.accepts_index_arg:
                return formatter(value, index=idx)
            return formatter(value)

        return {self.name: value}

    def add_child(self, child_tree: "Tree") -> None:
        self.children.append(child_tree)

    def get_child(self, path: str) -> "Tree":
        paths = path.split(".")

        def _select_child(tree: "Tree", path: str) -> "Tree":
            return next(filter(lambda p: p.path == path, tree.children))

        child = reduce(lambda tree, path: _select_child(tree, path), paths, self)
        return child

    def __repr__(self) -> str:
        base = f"{self.path}\n"

        for child in self.children:
            sep = "\t" * self.depth
            base += f"{sep}{repr(child)}\n"

        return base


def load_tree(data: dict, path: str, parent: Union[Tree, None] = None, root_name: str = "root") -> Tree:
    if parent is None:
        tree = Tree(parent, root_name)
    else:
        tree = Tree(parent=parent, path=path)

    if isinstance(data, dict):
        for k, v in data.items():
            child = load_tree(v, k, tree)
            tree.add_child(child)

    return tree


def set(obj: dict, path: str, value: Any) -> None:
    *split_path, last = path.split(".")
    for bit in split_path:
        obj = obj.setdefault(bit, {})
    obj[last] = value


def load_tree_rep(paths: Sequence[Union[str, Tuple[str, ColumnSpec]]]) -> Tree:
    tree_dict: Dict[str, Any] = {}
    for path in paths:
        if isinstance(path, tuple):
            path = path[0]
        set(tree_dict, path, None)
    tree = load_tree(tree_dict, "None", parent=None)

    specs: List[Tuple[str, ColumnSpec]] = [path for path in paths if isinstance(path, tuple)]

    for path, spec in specs:
        child = tree.get_child(path)
        child.tree_spec = spec
        formatter = spec.get("formatter")
        if formatter:
            argspec = inspect.getfullargspec(formatter)
            child.accepts_index_arg = "index" in argspec.args

    return tree



from typing import Any, Dict, List, Sequence, Union
from .types import ColumnDef
from .tree import Tree, load_tree_rep

__all__ = ["unwind"]


def _unwind(data: dict, tree: Tree) -> Union[List, List[Dict[str, Any]], Dict[str, Any], dict, List[dict]]:
    if not tree.children:
        tree_data = data.get(tree.path)
        if isinstance(data, list):
            return [_unwind(d, tree) for d in data]
        if isinstance(tree_data, list):
            return [tree.get_value(x, idx) for idx, x in enumerate(tree_data)]
        return tree.get_value(data.get(tree.path))

    level_list: List[dict] = []
    level_dict: dict = {}

    tree_data = data.get(tree.path)
    tree_data = {} if tree_data is None else tree_data
    if isinstance(tree_data, list):
        for item in tree_data:
            item_output_dict = {}
            item_accum: List[dict] = []
            for child in tree.children:
                value: Union[dict, list] = _unwind(item, child)
                if isinstance(value, list):
                    if item_accum and len(item_accum) == len(value):  # Handle adjacents
                        item_accum = [{**a, **b} for a, b in zip(item_accum, value)]
                    else:
                        item_accum.extend(value)
                else:
                    item_output_dict.update(value)
            if len(item_accum) > 0:
                level_list.extend([{**item_output_dict, **item} for item in item_accum])
            else:
                level_list.append(item_output_dict)
    else:
        for child in tree.children:
            value = _unwind(tree_data, child)
            if isinstance(value, list):
                level_list.extend(value)
            else:
                level_dict.update(value)

    if len(level_list) == 0:
        return level_dict

    appended = [{**level_dict, **list_value} for list_value in level_list]
    return appended


def unwind(data: Dict[Any, Any], columns: Sequence[ColumnDef]) -> List[Dict[str, Any]]:
    paths_list = list(columns)
    tree = load_tree_rep(paths_list)
    unwound = _unwind({"root": data}, tree)
    if not isinstance(unwound, list):
        return [unwound]

    return unwound



from typing import Any, Callable, Protocol, Tuple, TypedDict, Union, runtime_checkable
import sys

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict  # noqa: F811
else:
    from typing import TypedDict

__all__ = ["ColumnDef", "ColumnFormatter", "ColumnSpec"]


@runtime_checkable
class IndexFormatter(Protocol):
    def __call__(self, data: Any, index: int = ...):
        ...  # pragma: no cover


ColumnFormatter = Union[IndexFormatter, Callable[[Any], Any]]


class ColumnSpec(TypedDict, total=False):
    name: str
    formatter: ColumnFormatter


ColumnDef = Union[Tuple[str, ColumnSpec], str]
